#ifndef __GC_CARD_H__
#define __GC_CARD_H__

#define CARD_FILENAME     32
#define CARD_MAXFILES    127
#define CARD_MAXICONS      8
#define CARD_READSIZE    512
#define CARD_SECTORSIZE 8192
#define CARD_SLOTA         0
#define CARD_SLOTB         1
#define CARD_SYSAREA       5
#define CARD_WRITESIZE   128

/* Banner & Icon defines */

#define CARD_BANNER_NONE  0
#define CARD_BANNER_CI    1
#define CARD_BANNER_RGB   2
#define CARD_BANNER_W    96
#define CARD_BANNER_H    32

#define CARD_ICON_NONE    0
#define CARD_ICON_CI      1
#define CARD_ICON_RGB     2
#define CARD_ICON_LOOP    0
#define CARD_ICON_BOUNCE  4
#define CARD_ICON_W      32
#define CARD_ICON_H      32

#define CARD_SPEED_END 0
#define CARD_SPEED_4   1
#define CARD_SPEED_8   2
#define CARD_SPEED_12  3

/* File attribute defines */

#define CARD_ATTRIB_PUBLIC 0x04
#define CARD_ATTRIB_NOCOPY 0x08
#define CARD_ATTRIB_NOMOVE 0x10

/* Errors */

#define CARD_ERROR_NONE     0
#define CARD_ERROR_NOCARD  -1
#define CARD_ERROR_CORRUPT -2
#define CARD_ERROR_FATAL   -3
#define CARD_ERROR_INIT    -4
#define CARD_ERROR_NOSPACE -5
#define CARD_ERROR_NOENTRY -6
#define CARD_ERROR_EXISTS  -7

typedef struct {

  unsigned char gamecode[4];         // Game code
  unsigned char company[2];          // Company code
  unsigned char padding_ff;          // Always 0xff
  unsigned char banner_fmt;          // Banner format
  unsigned char filename[CARD_FILENAME];  // Filename
  unsigned long time;                // Time of last modification
  unsigned long icon_addr;           // Offset of image data (must be 0-511)
  unsigned short icon_fmt;           // Icon formats
  unsigned short icon_speed;         // Icon speeds
  unsigned char attrib;              // File attribute
  char padding_00;                   // Always 0x00
  unsigned short block;              // Block/sector file starts at
  unsigned short length;             // Length in blocks/sectors
  unsigned short padding_ffff;       // Always 0xffff;
  unsigned long comment_addr;        // Offset of comment 
} GC_Card_DirEntry;

typedef struct {

  GC_Card_DirEntry entries[CARD_MAXFILES];    // Entries for files
  unsigned char padding[58];                  // Unused bytes after last entry
  unsigned short num;                         // Increases with each update
  unsigned short checksum1;                   // 1st part of checksum
  unsigned short checksum2;                   // 2nd part of checksum
} GC_Card_DirBlock;

typedef struct {

  unsigned short checksum1;                   // 1st part of checksum
  unsigned short checksum2;                   // 2nd part of checksum
  unsigned short num;                         // Increases with each update
  unsigned short freeblocks;                  // Amount of free blocks
  unsigned short last_alloc;                  // Last allocated block
  unsigned short fat[0xffb];                  // Table of allocated blocks
} GC_Card_FatBlock;

typedef struct {

  unsigned char gamecode[5];         // Game code
  unsigned char company[3];          // Company code
  unsigned char banner_fmt;          // Banner format
  unsigned char filename[CARD_FILENAME+1]; // Filename
  unsigned long time;                // Time of last modification
  unsigned long icon_addr;           // Offset of image data (must be 0-511)
  unsigned short icon_fmt;           // Icon formats
  unsigned short icon_speed;         // Icon speeds
  unsigned char attrib;              // File attribute
  unsigned long comment_addr;        // Offset of comment
  unsigned long size;                // Size in bytes
  int filenum;                       // File number
} GC_Card_File;

typedef struct {

  unsigned long banner;                       // Offset to banner data
  unsigned long banner_tlut;                  // Offset to banner's TLUT
  unsigned long icons[CARD_MAXICONS];         // Offsets for the icons
  unsigned long icons_tlut;                   // Offset of the icons' TLUT
  unsigned long data;                         // Offset to the save data
} GC_Card_Offsets;


/* Get icon attribute macros */

#define GC_Card_GetIconFmt(cfile, icon)   ((cfile->icon_fmt >> (icon * 2)) & 3)
#define GC_Card_GetIconSpeed(cfile, icon) ((cfile->icon_speed >> (icon * 2)) & 3)
#define GC_Card_GetIconAnim(cfile)        (cfile->banner_fmt & 4)

unsigned short GC_Card_RetrieveID (unsigned long channel);
unsigned char GC_Card_ReadStatus (unsigned long channel);
void GC_Card_ClearStatus (unsigned long channel);
void GC_Card_ChipErase (unsigned long channel);
void GC_Card_SectorErase (unsigned long channel, unsigned long sector);
void GC_Card_SectorProgram (unsigned long channel, unsigned char *abuf, unsigned long address, unsigned long size);
void GC_Card_ReadArray (unsigned long channel, unsigned char *abuf, unsigned long address, unsigned long size);
int GC_Card_IsPresent(unsigned long channel);
int GC_Card_Begin(unsigned long channel, int *fixed);
int GC_Card_GetOffsets(GC_Card_File *cfile, GC_Card_Offsets *offsets);
int GC_Card_CreateFile(GC_Card_File *cfile, GC_Card_Offsets *offsets);
int GC_Card_OpenFile(GC_Card_File *cfile, GC_Card_Offsets *offsets);
int GC_Card_OpenFilenum(GC_Card_File *cfile, GC_Card_Offsets *offsets, int filenum);
int GC_Card_WriteFile(GC_Card_File *cfile, void *buf);
int GC_Card_ReadFile(GC_Card_File *cfile, void *buf);
int GC_Card_DeleteFile(GC_Card_File *cfile);
void GC_Card_SetIcon(GC_Card_File *cfile, int icon, int fmt, int speed);
void GC_Card_SetIconAnim(GC_Card_File *cfile, int anim);
int GC_Card_CalcSize(GC_Card_File *cfile, int datalen);
int GC_Card_FreeSpace();
int GC_Card_DrawIcon(GC_Card_File *cfile, GC_Card_Offsets *offsets, int icon, void *buf, unsigned long *fb, int x, int y);

#endif
