// GCLIB (Free GameCube SDK Libraries)
//
// GC_Card.c by Costis and CrowTRobo
//
// Version 1.0
//
// Revisions:
//
// V1.0 - I have written the low-level memory card
//		  accessing functions, and CrowTRobo has
//		  written all of the high-level file system
//	      code.

#include <GC_EXI.h>
#include <GC_Card.h>
#include <GC_Memory.h>
#include <GC_Time.h>
#include <GC_Video.h>

GC_Card_DirBlock dirblock[2] __attribute__ ((aligned (32)));
GC_Card_FatBlock fatblock[2] __attribute__ ((aligned (32)));
unsigned char sysarea[CARD_SYSAREA*CARD_SECTORSIZE] __attribute__ ((aligned (32)));
unsigned long curslot, numsectors;
int card_init = 0, curdir, curfat;

unsigned short GC_Card_RetrieveID (unsigned long channel)
{
	int i;
  unsigned char cbuf[4];

	// Select the specified EXI channel and device.
	GC_EXI_Select (channel, 0, 4);
			
	// Send the EXI ID command (0x85xx)
	cbuf[0] = 0x85; // Command Byte
	cbuf[1] = 0x00;
	GC_EXI_Imm (channel, cbuf, 2, EXI_CONTROL_TYPE_WRITE);

	// Read the actual ID data (2 bytes)
	GC_EXI_Imm (channel, cbuf, 2, EXI_CONTROL_TYPE_READ);

	// Deselect the selected EXI device.
	GC_EXI_Deselect (channel);

	return *(unsigned short*)cbuf;
}

unsigned char GC_Card_ReadStatus (unsigned long channel)
{
  unsigned char cbuf[4];

	// Select the specified EXI channel and device.
	GC_EXI_Select (channel, 0, 4);
			
	// Send the EXI ID command (0x83xx)
	cbuf[0] = 0x83; // Command Byte
	cbuf[1] = 0x00;
	GC_EXI_Imm (channel, cbuf, 2, EXI_CONTROL_TYPE_WRITE);

	// Read the actual ID data (2 bytes)
	GC_EXI_Imm (channel, cbuf, 1, EXI_CONTROL_TYPE_READ);

	// Deselect the selected EXI device.
	GC_EXI_Deselect (channel);

	return cbuf[0];
}

void GC_Card_ClearStatus (unsigned long channel)
{
	int i;
  unsigned char cbuf[4];

	// Select the specified EXI channel and device.
	GC_EXI_Select (channel, 0, 4);
			
	// Send the EXI ID command (0x89)
	cbuf[0] = 0x89; // Command Byte
	GC_EXI_Imm (channel, cbuf, 1, EXI_CONTROL_TYPE_WRITE);

	// Deselect the selected EXI device.
	GC_EXI_Deselect (channel);
}

void GC_Card_ChipErase (unsigned long channel)
{
	int i;
  unsigned char cbuf[3];

	// Select the specified EXI channel and device.
	GC_EXI_Select (channel, 0, 4);

	// Send the EXI Chip Erase command (0xF4xxxx)
	cbuf[0] = 0xF4; // Command Byte
	cbuf[1] = 0x00;
	cbuf[2] = 0x00;
	GC_EXI_Imm (channel, cbuf, 3, EXI_CONTROL_TYPE_WRITE);

	// Deselect the selected EXI device.
	GC_EXI_Deselect (channel);
}

/*
  CrowTRobo changes:
  Checked status to see if card is ready for a command and waits till
  the erase is finished.
*/

void GC_Card_SectorErase (unsigned long channel, unsigned long sector)
{
	int i;
  unsigned char cbuf[3];

  // Check if the card is ready
  while (!(GC_Card_ReadStatus(channel) & 1));

	// Select the specified EXI channel and device.
	GC_EXI_Select (channel, 0, 4);

	// Send the EXI Sector Erase command (0xF1xxxx)
	cbuf[0] = 0xF1; // Command Byte
	cbuf[1] = (sector >> 17) & 0x7F;
	cbuf[2] = (sector >> 9) & 0xFF;
	GC_EXI_Imm (channel, cbuf, 3, EXI_CONTROL_TYPE_WRITE);

	// Deselect the selected EXI device.
	GC_EXI_Deselect (channel);

  // Wait till the erase is finished
  while (GC_Card_ReadStatus(channel) & 0x8000);
}

/*
  CrowTRobo changes:
  Data must be written 128 bytes at a time, so implemented that.
  Used EXI_Dma instead of writing one byte at a time.
  Checked status to see if card is ready for a command and waits till
  the write is finished.
*/

void GC_Card_SectorProgram (unsigned long channel, unsigned char *abuf, unsigned long address, unsigned long size)
{
	int i;
  unsigned char cbuf[4];
	unsigned char *bbuf = abuf;

  for (i = 0; i < size / CARD_WRITESIZE; i++) {
          
    // Check if the card is ready
    while (!(GC_Card_ReadStatus(channel) & 1));

    // Select the specified EXI channel and device.
    GC_EXI_Select (channel, 0, 4);

    // Send the EXI Sector Program command (0xF2xxxxxx)
    cbuf[0] = 0xF2; // Command Byte
    cbuf[1] = (address >> 17) & 0x3F;
    cbuf[2] = (address >> 9) & 0xFF;
    cbuf[3] = (address >> 7) & 3;
    GC_EXI_Imm (channel, cbuf, 4, EXI_CONTROL_TYPE_WRITE);

    cbuf[0] = address & 0x7F;
    GC_EXI_Imm (channel, cbuf, 1, EXI_CONTROL_TYPE_WRITE);

    GC_EXI_Dma(channel, bbuf, CARD_WRITESIZE, EXI_CONTROL_TYPE_WRITE);

    // Deselect the selected EXI device.
    GC_EXI_Deselect (channel);

    // Wait till the write is finished
    while (GC_Card_ReadStatus(channel) & 0x8000);

    address += CARD_WRITESIZE;
    bbuf += CARD_WRITESIZE;
  }
}

/*
  CrowTRobo changes:
  Data must be read 512 bytes at a time, so implemented that.
  Used EXI_Dma instead of reading one byte at a time.
  Checked status to see if the card is ready for a command.
*/

void GC_Card_ReadArray (unsigned long channel, unsigned char *abuf, unsigned long address, unsigned long size)
{
	int i;
  unsigned char cbuf[4];
  unsigned char *bbuf = abuf;

  for (i = 0; i < size / CARD_READSIZE; i++) {

    // Check if the card is ready
    while (!(GC_Card_ReadStatus(channel) & 1));

    // Select the specified EXI channel and device.
    GC_EXI_Select (channel, 0, 4);

    // Send the EXI Sector Read command (0x52xxxxxx)
    cbuf[0] = 0x52; // Command Byte
    cbuf[1] = (address >> 17) & 0x3F;
    cbuf[2] = (address >> 9) & 0xFF;
    cbuf[3] = (address >> 7) & 3;
    GC_EXI_Imm (channel, cbuf, 4, EXI_CONTROL_TYPE_WRITE);

    cbuf[0] = address & 0x7F;
    GC_EXI_Imm (channel, cbuf, 1, EXI_CONTROL_TYPE_WRITE);

    cbuf[0] = 0;
    cbuf[1] = 0;
    cbuf[2] = 0;
    cbuf[3] = 0;
    GC_EXI_Imm (channel, cbuf, 4, EXI_CONTROL_TYPE_WRITE);

    GC_EXI_Dma(channel, bbuf, CARD_READSIZE, EXI_CONTROL_TYPE_READ);

    // Deselect the selected EXI device.
    GC_EXI_Deselect (channel);

    address += CARD_READSIZE;
    bbuf += CARD_READSIZE;
  }
}

/*
  Determines if a memcard is present in the given slot.
*/

int GC_Card_IsPresent(unsigned long channel) {

  unsigned long id;

  id = GC_EXI_RetrieveID(channel, 0);

  if (id & 0xffff0000 || id & 3)
    return 0;

  return id;
}

/*
  Calculates checksum used by the dir and fat blocks.
*/

void GC_Card_Checksum(unsigned short *buf, int len, unsigned short *cs1, unsigned short *cs2) {

  unsigned short val;
  int i;

  *cs1 = 0;
  *cs2 = 0;
  len /= 2;

  for (i = 0; i < len; i++) {
    val = *cs1;
    *cs1 = buf[i] + val;

    val = *cs2;
    *cs2 = ~buf[i] + val;
  }

  if (*cs1 == 0xffff)
    *cs1 = 0;

  if (*cs2 == 0xffff)
    *cs2 = 0;
}

/*
  Updates the dir block on the memory card.
*/

void GC_Card_UpdateDir(unsigned long channel) {

  dirblock[curdir].num++;

  // Checksum, erase and write current dir block

  GC_Card_Checksum((unsigned short *)&dirblock[curdir], 0x1ffc, &dirblock[curdir].checksum1, &dirblock[curdir].checksum2);
  GC_Memory_DCFlushRange(&dirblock[curdir], sizeof(GC_Card_DirBlock));
  GC_Card_SectorErase(channel, curdir*0x2000+0x2000);
  GC_Card_SectorProgram(channel, (unsigned char *)&dirblock[curdir], curdir*0x2000+0x2000, 0x2000);

  // Swap dir blocks

  curdir ^= 1;
  GC_Memory_memcpy(&dirblock[curdir], &dirblock[curdir^1], sizeof(GC_Card_DirBlock));
}

/*
  Updates the fat block on the memory card.
*/

void GC_Card_UpdateFat(unsigned long channel) {

  fatblock[curfat].num++;

  // Checksum, erase and write current fat block

  GC_Card_Checksum((unsigned short *)&fatblock[curfat]+2, 0x1ffc, &fatblock[curfat].checksum1, &fatblock[curfat].checksum2);
  GC_Memory_DCFlushRange(&fatblock[curfat], sizeof(GC_Card_FatBlock));
  GC_Card_SectorErase(channel, curfat*0x2000+0x6000);
  GC_Card_SectorProgram(channel, (unsigned char *)&fatblock[curfat], curfat*0x2000+0x6000, 0x2000);

  // Swap fat blocks

  curfat ^= 1;
  GC_Memory_memcpy(&fatblock[curfat], &fatblock[curfat^1], sizeof(GC_Card_FatBlock));
}

/*
  Checks to see if the dir and fat blocks are valid, fixes bad ones
  if it can and sets up which dir block and fat block to use.
*/

int GC_Card_Check(unsigned long channel, int *fixed) {

  unsigned short checksum1, checksum2;
  int i, bad = 0;

  *fixed = 0;

  for (i = 0; i < 2; i++) {
    GC_Card_Checksum((unsigned short *)&dirblock[i], 0x1ffc, &checksum1, &checksum2);

    if (checksum1 != dirblock[i].checksum1 || checksum2 != dirblock[i].checksum2)
      bad |= (i + 1);
  }

  if (bad & 3) {

    // If both dir blocks are bad, then the card is beyond repair.

    if (bad == 3)
      return CARD_ERROR_CORRUPT;

    // If only one block is bad, then copy the good block over it.

    if (bad == 1) {
      GC_Memory_memcpy(&dirblock[0], &dirblock[1], sizeof(GC_Card_DirBlock));
      curdir = 0;
    }
    else {
      GC_Memory_memcpy(&dirblock[1], &dirblock[0], sizeof(GC_Card_DirBlock));
      curdir = 1;
    }

    *fixed |= 1;
  }
  else {

    // No errors, so use the older block and update it.

    if (dirblock[0].num < dirblock[1].num) {
      GC_Memory_memcpy(&dirblock[0], &dirblock[1], sizeof(GC_Card_DirBlock));
      curdir = 0;
    }
    else {
      GC_Memory_memcpy(&dirblock[1], &dirblock[0], sizeof(GC_Card_DirBlock));
      curdir = 1;
    }
  }

  bad = 0;

  for (i = 0; i < 2; i++) {
    GC_Card_Checksum((unsigned short *)&fatblock[i]+2, 0x1ffc, &checksum1, &checksum2);

    if (checksum1 != fatblock[i].checksum1 || checksum2 != fatblock[i].checksum2)
      bad |= (i + 1);
  }

  if (bad & 3) {

    // If both fat blocks are bad, then the card is beyond repair.

    if (bad == 3)
      return CARD_ERROR_CORRUPT;

    // If only one block is bad, then copy the good block over it.

    if (bad == 1) {
      GC_Memory_memcpy(&fatblock[0], &fatblock[1], sizeof(GC_Card_FatBlock));
      curfat = 0;
    }
    else {
      GC_Memory_memcpy(&fatblock[1], &fatblock[0], sizeof(GC_Card_FatBlock));
      curfat = 1;
    }

    *fixed |= 2;
  }
  else {

    // No errors, so use the older block and update it.

    if (fatblock[0].num < fatblock[1].num) {
      GC_Memory_memcpy(&fatblock[0], &fatblock[1], sizeof(GC_Card_FatBlock));
      curfat = 0;
    }
    else {
      GC_Memory_memcpy(&fatblock[1], &fatblock[0], sizeof(GC_Card_FatBlock));
      curfat = 1;
    }
  }

  return CARD_ERROR_NONE;
}

/*
  Calculates the banner and icon offsets for a given file.
*/

int GC_Card_GetOffsets(GC_Card_File *cfile, GC_Card_Offsets *offsets) {

  unsigned long addr, bsize, isize;
  int i, format, needs_tlut = 0;

  if (card_init != 1)
    return CARD_ERROR_INIT;

  if (cfile->filenum < 0 || cfile->filenum >= CARD_MAXFILES)
    return CARD_ERROR_FATAL;

  addr = cfile->icon_addr;
  bsize = CARD_BANNER_W * CARD_BANNER_H;
  isize = CARD_ICON_W * CARD_ICON_H;

  // The banner is the first data at the icon address.

  if (cfile->banner_fmt & CARD_BANNER_CI) {

    // Color Index format = 8bit per pixel and 256x16bit palette.

    offsets->banner = addr;
    offsets->banner_tlut = addr + bsize;
    addr += bsize + 512;
  }
  else if (cfile->banner_fmt & CARD_BANNER_RGB) {

    // RGB5/RGB4A3 = 16bits per pixel, no palette.

    offsets->banner = addr;
    offsets->banner_tlut = 0xffff;
    addr += bsize * 2;
  }
  else {
    offsets->banner = 0xffff;
    offsets->banner_tlut = 0xffff;
  }

  for (i = 0; i < CARD_MAXICONS; i++) {

    // icon_fmt has 2 bits per icon to describe format of each one

    format = cfile->icon_fmt >> (i * 2);

    if (format & CARD_ICON_CI) {
      offsets->icons[i] = addr;
      addr += isize;
      needs_tlut = 1;
    }
    else if (format & CARD_ICON_RGB) {
      offsets->icons[i] = addr;
      addr += isize * 2;
    }
    else 
      offsets->icons[i] = 0xffff;      
  }

  if (needs_tlut) {
    offsets->icons_tlut = addr;
    addr += 512;
  }
  else
    offsets->icons_tlut = 0xffff;

  offsets->data = addr;

  return CARD_ERROR_NONE;
}

/*
  Initializes everything in order to begin using a memcard.
*/

int GC_Card_Begin(unsigned long channel, int *fixed) {

  int i, id, ret;

  card_init = 0;

  if (channel < 0 || channel >= 2)
    return CARD_ERROR_FATAL;

  id = GC_Card_IsPresent(channel);

  if (id == 0)
    return CARD_ERROR_NOCARD;

  GC_Memory_DCInvalidateRange(sysarea, CARD_SYSAREA*CARD_SECTORSIZE);
  GC_Card_ReadArray(channel, sysarea, 0, CARD_SYSAREA*CARD_SECTORSIZE);
  GC_Memory_DCFlushRange(sysarea, CARD_SYSAREA*CARD_SECTORSIZE);

  // Load the dirblock structures 

  GC_Memory_memcpy(&dirblock[0], sysarea+0x2000, sizeof(GC_Card_DirBlock));
  GC_Memory_memcpy(&dirblock[1], sysarea+0x4000, sizeof(GC_Card_DirBlock));
  GC_Memory_DCFlushRange(dirblock, sizeof(GC_Card_DirBlock)*2);

  // Load the fatblock structures

  GC_Memory_memcpy(&fatblock[0], sysarea+0x6000, sizeof(GC_Card_FatBlock));
  GC_Memory_memcpy(&fatblock[1], sysarea+0x8000, sizeof(GC_Card_FatBlock));
  GC_Memory_DCFlushRange(fatblock, sizeof(GC_Card_FatBlock)*2);

  ret = GC_Card_Check(channel, fixed);

  if (ret != CARD_ERROR_NONE)
    return ret;

  // If the dir or fat block had to be repaired, update them.

  if (*fixed & 1)
    GC_Card_UpdateDir(channel);

  if (*fixed & 2)
    GC_Card_UpdateFat(channel);   

  card_init = 1;
  curslot = channel;
  numsectors = id << 4;

  return CARD_ERROR_NONE;
}

/*
  Creates a file on the memcard.
*/

int GC_Card_CreateFile(GC_Card_File *cfile, GC_Card_Offsets *offsets) {
                                         
  GC_Card_DirEntry *entry;
  unsigned short block, prevblock, curblock;
  int i, count, filenum = -1;

  cfile->filenum = -1;

  if (card_init != 1)
    return CARD_ERROR_INIT;

  if (GC_Memory_strlen(cfile->filename) > CARD_FILENAME ||
      GC_Memory_strlen(cfile->gamecode) != 4 ||
      GC_Memory_strlen(cfile->company) != 2) {

    return CARD_ERROR_FATAL;
  }

  // size must be a multiple of the sector size

  if (cfile->size == 0 || cfile->size % CARD_SECTORSIZE)
    return CARD_ERROR_FATAL;

  if ((fatblock[curfat].freeblocks * CARD_SECTORSIZE) < cfile->size)
    return CARD_ERROR_NOSPACE;

  // Find a free entry and check to see if the file already exists.

  entry = dirblock[curdir].entries;

  for (i = 0; i < CARD_MAXFILES; i++) {

    if (entry[i].gamecode[0] == 0xff) {

      if (filenum == -1)
        filenum = i;
    }
    else {

      if (GC_Memory_memcmp(entry[i].gamecode, cfile->gamecode, 4) == 0 &&
          GC_Memory_memcmp(entry[i].company, cfile->company, 2) == 0 &&
          GC_Memory_memcmp(entry[i].filename, cfile->filename, CARD_FILENAME) == 0) {

        return CARD_ERROR_EXISTS;
      }
    }
  }

  if (filenum == -1)
    return CARD_ERROR_NOENTRY;

  // Add which blocks this file will take up into the FAT

  block = 0xffff;
  curblock = fatblock[curfat].last_alloc;
  i = cfile->size / CARD_SECTORSIZE;
  count = 0;

  while (1) {

    if (i == 0) {

      // Done allocating blocks

      fatblock[curfat].freeblocks -= cfile->size / CARD_SECTORSIZE;
      fatblock[curfat].last_alloc = curblock;
      GC_Card_UpdateFat(curslot);
      break;
    }

    /*
      Since testing free space has already been done, if all the blocks
      the file takes up cannot be entered into the FAT, something is
      wrong.
    */

    count++;

    if (count >= (numsectors - CARD_SYSAREA))
      return CARD_ERROR_CORRUPT;

    curblock++;

    if (curblock < 5 || curblock >= numsectors)
      curblock = 5;

    if (fatblock[curfat].fat[curblock-5] == 0) {

      if (block != 0xffff)
        fatblock[curfat].fat[prevblock-5] = curblock;
      else
        block = curblock;

      fatblock[curfat].fat[curblock-5] = 0xffff;
      prevblock = curblock;
      i--;
    }
  }

  // Finally, copy the necessary info to the file entry in the dir block.

  entry = &dirblock[curdir].entries[filenum];

  GC_Memory_memcpy(entry->filename, cfile->filename, CARD_FILENAME);
  GC_Memory_memcpy(entry->gamecode, cfile->gamecode, 4);
  GC_Memory_memcpy(entry->company, cfile->company, 2);

  entry->padding_ff = 0xff;                            
  entry->banner_fmt = cfile->banner_fmt;                               
  entry->time = GC_Time_GetTime() / (*(volatile unsigned long *)(0x800000f8) / 4);                                     
  entry->icon_addr = cfile->icon_addr;                          
  entry->icon_fmt = cfile->icon_fmt;                                
  entry->icon_speed = cfile->icon_speed;
  entry->attrib = cfile->attrib;

  if (entry->attrib == 0)
    entry->attrib = CARD_ATTRIB_PUBLIC;

  entry->padding_00 = 0;                              
  entry->block = block;                               
  entry->length = cfile->size / CARD_SECTORSIZE;             
  entry->padding_ffff = 0xffff;                        
  entry->comment_addr = cfile->comment_addr;                        
  cfile->filenum = filenum;
  cfile->time = entry->time;
  
  GC_Card_UpdateDir(curslot);
  GC_Card_GetOffsets(cfile, offsets);
  
  return CARD_ERROR_NONE;
}

/*
  Opens a memcard file.
*/

int GC_Card_OpenFile(GC_Card_File *cfile, GC_Card_Offsets *offsets) {

  GC_Card_DirEntry *entry;
  int i;

  cfile->filenum = -1;

  if (card_init != 1)
    return CARD_ERROR_INIT;

  if (GC_Memory_strlen(cfile->filename) > CARD_FILENAME ||
      GC_Memory_strlen(cfile->gamecode) != 4 ||
      GC_Memory_strlen(cfile->company) != 2)
    return CARD_ERROR_FATAL;

  entry = dirblock[curdir].entries;

  for (i = 0; i < CARD_MAXFILES; i++) {

    if (entry[i].gamecode[0] != 0xff) {

      if (GC_Memory_memcmp(entry[i].gamecode, cfile->gamecode, 4) == 0 &&
          GC_Memory_memcmp(entry[i].company, cfile->company, 2) == 0 &&
          GC_Memory_memcmp(entry[i].filename, cfile->filename, CARD_FILENAME) == 0) {

        break;
      }
    }
  }

  // If the for loop ran it's course, then no file matches.

  if (i >= CARD_MAXFILES)
    return CARD_ERROR_NOENTRY;

  // File exists, so fill in the supplied structures.

  GC_Memory_memcpy(cfile->filename, entry[i].filename, CARD_FILENAME);
  GC_Memory_memcpy(cfile->gamecode, entry[i].gamecode, 4);
  GC_Memory_memcpy(cfile->company, entry[i].company, 2);
  cfile->filename[32] = '\0';
  cfile->gamecode[4] = '\0';
  cfile->company[2] = '\0';

  cfile->banner_fmt = entry[i].banner_fmt;
  cfile->time = entry[i].time;
  cfile->icon_addr = entry[i].icon_addr;
  cfile->icon_fmt = entry[i].icon_fmt;
  cfile->icon_speed = entry[i].icon_speed;
  cfile->attrib = entry[i].attrib;
  cfile->comment_addr = entry[i].comment_addr;
  cfile->size = entry[i].length * CARD_SECTORSIZE;
  cfile->filenum = i;

  GC_Card_GetOffsets(cfile, offsets);

  return CARD_ERROR_NONE;
}

/*
  Opens a file by file number.
*/

int GC_Card_OpenFilenum(GC_Card_File *cfile, GC_Card_Offsets *offsets, int filenum) {

  GC_Card_DirEntry *entry;

  if (card_init != 1)
    return CARD_ERROR_INIT;

  if (filenum < 0 || filenum >= CARD_MAXFILES)
    return CARD_ERROR_FATAL;

  entry = &dirblock[curdir].entries[filenum];

  // Just load the gamecode, company and filename from the given file
  // number and use the regular open function, which will take care of
  // seeing if the file actually exists, etc.

  GC_Memory_memcpy(cfile->gamecode, entry->gamecode, 4);
  GC_Memory_memcpy(cfile->company, entry->company, 2);
  GC_Memory_memcpy(cfile->filename, entry->filename, CARD_FILENAME);
  cfile->gamecode[4] = '\0';
  cfile->company[2] = '\0';
  cfile->filename[CARD_FILENAME] = '\0';

  return GC_Card_OpenFile(cfile, offsets);
}

/*
  Writes a file to the memcard.
*/

int GC_Card_WriteFile(GC_Card_File *cfile, void *buf) {

  GC_Card_DirEntry *entry;
  unsigned short block;
  int i;

  if (card_init != 1)
    return CARD_ERROR_INIT;

  if (cfile->filenum < 0 || cfile->filenum >= CARD_MAXFILES)
    return CARD_ERROR_FATAL;

  entry = &dirblock[curdir].entries[cfile->filenum];
  block = entry->block;
  GC_Memory_DCFlushRange(buf, cfile->size);

  for (i = 0; i < entry->length; i++) {

    if (block < CARD_SYSAREA || block >= numsectors)
      return CARD_ERROR_CORRUPT;

    GC_Card_SectorErase(curslot, block*CARD_SECTORSIZE);
    GC_Card_SectorProgram(curslot, buf+i*CARD_SECTORSIZE, block*CARD_SECTORSIZE, CARD_SECTORSIZE);

    // Look up the next block the files takes up in the fat

    block = fatblock[curfat].fat[block-5];
  }

  // Update the file's dir entry.

  GC_Memory_memcpy(entry->filename, cfile->filename, CARD_FILENAME);
  GC_Memory_memcpy(entry->gamecode, cfile->gamecode, 4);
  GC_Memory_memcpy(entry->company, cfile->company, 2);

  entry->banner_fmt = cfile->banner_fmt;                               

  entry->time = *(volatile unsigned long *)(0x800000f8) / 4;
  entry->time = GC_Time_GetTime() / entry->time;

  entry->icon_addr = cfile->icon_addr;
  entry->icon_fmt = cfile->icon_fmt;                                
  entry->icon_speed = cfile->icon_speed;                              
  entry->attrib = cfile->attrib;                                  
  entry->comment_addr = cfile->comment_addr;                        
  cfile->time = entry->time;

  GC_Card_UpdateDir(curslot);
  
  return CARD_ERROR_NONE;
}
    
/*
  Reads a file from the memcard.
*/

int GC_Card_ReadFile(GC_Card_File *cfile, void *buf) {

  GC_Card_DirEntry *entry;
  unsigned short block;
  int i;

  if (card_init != 1)
    return CARD_ERROR_INIT;

  if (cfile->filenum < 0 || cfile->filenum >= CARD_MAXFILES)
    return CARD_ERROR_FATAL;

  entry = &dirblock[curdir].entries[cfile->filenum];
  block = entry->block;
  GC_Memory_DCInvalidateRange(buf, cfile->size);

  for (i = 0; i < entry->length; i++) {

    if (block < CARD_SYSAREA || block >= numsectors)
      return CARD_ERROR_CORRUPT;

    GC_Card_ReadArray(curslot, buf+i*CARD_SECTORSIZE, block*CARD_SECTORSIZE, CARD_SECTORSIZE);

    // Look up the next block the files takes up in the fat

    block = fatblock[curfat].fat[block-5];
  }

  GC_Memory_DCFlushRange(buf, cfile->size);

  return CARD_ERROR_NONE;
}

/*
  Deletes a file from the memcard.
*/

int GC_Card_DeleteFile(GC_Card_File *cfile) {

  unsigned short block, prevblock;

  if (card_init != 1)
    return CARD_ERROR_INIT;

  if (cfile->filenum < 0 || cfile->filenum >= CARD_MAXFILES)
    return CARD_ERROR_FATAL;

  // Get the file's initial block

  block = dirblock[curdir].entries[cfile->filenum].block;

  while (1) {

    if (block == 0xffff)
      break;

    if (block < CARD_SYSAREA || block >= numsectors)
      return CARD_ERROR_CORRUPT;

    // Get the file's next block and clear the previous one from the fat

    prevblock = block;
    block = fatblock[curfat].fat[prevblock-5];
    fatblock[curfat].fat[prevblock-5] = 0;

    fatblock[curfat].freeblocks++;
  }

  GC_Card_UpdateFat(curslot);

  GC_Memory_memset(&dirblock[curdir].entries[cfile->filenum], 0xff, sizeof(GC_Card_DirEntry));
  GC_Card_UpdateDir(curslot);
  cfile->filenum = -1;

  return CARD_ERROR_NONE;
}

/*
  Sets an icon's attributes.
*/

void GC_Card_SetIcon(GC_Card_File *cfile, int icon, int fmt, int speed) {

  // Clear the two format bits of this icon

  cfile->icon_fmt &= ~(3 << (icon * 2));

  // Set the two format bits

  cfile->icon_fmt |= (fmt << (icon * 2));

  // Do the same for the two speed bits.

  cfile->icon_speed &= ~(3 << (icon * 2));
  cfile->icon_speed |= (speed << (icon * 2));
}

/*
  Sets the icon animation type for a file.
*/

void GC_Card_SetIconAnim(GC_Card_File *cfile, int anim) {

  // The animation is set in the third bit of 'banner_fmt'

  cfile->banner_fmt &= ~4;
  cfile->banner_fmt |= anim;
}

/*
  Calculates the needed file size based on banner, icons and save data.
*/

int GC_Card_CalcSize(GC_Card_File *cfile, int datalen) {

  GC_Card_Offsets offsets;
  int ret;

  ret = GC_Card_GetOffsets(cfile, &offsets);

  if (ret != CARD_ERROR_NONE)
    return ret;

  cfile->size = offsets.data + datalen;
  cfile->size += (CARD_SECTORSIZE - (cfile->size % CARD_SECTORSIZE));

  return CARD_ERROR_NONE;
}

/*
  Returns the amount of free space, in bytes, left on the memory card.
*/

int GC_Card_FreeSpace() {

  if (card_init != 1)
    return CARD_ERROR_INIT;

  return fatblock[curfat].freeblocks * CARD_SECTORSIZE;
}

/*
  Converts a packed RGB color to YCbCr.
*/

unsigned long GC_Card_ConvColor(unsigned short color) {

  unsigned char r, g, b;

  if (color & 0x8000) {

    // color is in RGB5 format

    r = (color >> 7) & 0xf8;
    g = (color >> 2) & 0xf8;
    b = (color << 3) & 0xf8;
  }
  else {

    //color is in RGB4A3 format

    r = (color >> 4) & 0xf0;
    g = color & 0xf0;
    b = (color << 4) & 0xf0;
  }

  return GC_Video_RGBToYCbCr(r, g, b);
}

/*
  Draws a RGB tile to the given framebuffer.
*/

void GC_Card_TileRGB(void *buf, unsigned long *fb, int x, int y, int tilenum) {

  unsigned char cb, cr;
  unsigned short *ptr;
  unsigned long clr1, clr2;
  int i, j;

  ptr = buf + tilenum * 32;

  for (i = y; i < y+4; i++) {

    for (j = x; j < x+2; j++) {
      clr1 = GC_Card_ConvColor(*ptr++);
      clr2 = GC_Card_ConvColor(*ptr++);

      cb = (((clr1 >> 16) & 0xff) + ((clr2 >> 16) & 0xff)) / 2;
      cr = ((clr1 & 0xff) + (clr2 & 0xff)) / 2;

      fb[i*320+j] = (clr1 & 0xff000000) | (cb << 16) | (clr2 & 0xff00) | cr;
    }
  }
}

/*
  Draws a C8 tile to the given framebuffer.
*/

void GC_Card_TileC8(void *buf, unsigned short *tlut, unsigned long *fb, int x, int y, int tilenum) {

  unsigned char *ptr, cb, cr;
  unsigned long clr1, clr2;
  int i, j;

  ptr = buf + tilenum * 32;

  for (i = y; i < y+4; i++) {

    for (j = x; j < x+4; j++) {
      clr1 = GC_Card_ConvColor(tlut[*ptr++]);
      clr2 = GC_Card_ConvColor(tlut[*ptr++]);

      cb = (((clr1 >> 16) & 0xff) + ((clr2 >> 16) & 0xff)) / 2;
      cr = ((clr1 & 0xff) + (clr2 & 0xff)) / 2;

      fb[i*320+j] = (clr1 & 0xff000000) | (cb << 16) | (clr2 & 0xff00) | cr;
    }
  }
}
  
/*
  Draws a file banner or icon to the given framebuffer.
*/

int GC_Card_DrawIcon(GC_Card_File *cfile, GC_Card_Offsets *offsets, int icon, void *buf, unsigned long *fb, int x, int y) {

  void *ptr;
  unsigned short *tlut;
  int i, j, k, w, h, tilew, tileh, format;

  if (card_init != 1)
    return CARD_ERROR_INIT;

  if (cfile->filenum < 0 || cfile->filenum >= CARD_MAXFILES)
    return CARD_ERROR_FATAL;

  if (icon < -1 || icon >= CARD_MAXICONS)
    return CARD_ERROR_FATAL;

  if (icon == -1) {
    w = CARD_BANNER_W;
    h = CARD_BANNER_H;
    ptr = buf + offsets->banner;

    // Assign format with icon defines to keep things consistent.

    if (cfile->banner_fmt & CARD_BANNER_CI) 
      format = CARD_ICON_CI;
    else if (cfile->banner_fmt & CARD_BANNER_RGB)
      format = CARD_ICON_RGB;
    else
      format = CARD_ICON_NONE;
  }
  else {
    w = CARD_ICON_W;
    h = CARD_ICON_H;
    ptr = buf + offsets->icons[icon];
    format = (cfile->icon_fmt >> (icon * 2)) & 3;
  }

  if (format == CARD_ICON_NONE)
    return CARD_ERROR_NOENTRY;
  else if (format == CARD_ICON_CI) {
    tilew = 8;
    tileh = 4;

    if (icon == -1) 
      tlut = buf + offsets->banner_tlut;
    else 
      tlut = buf + offsets->icons_tlut;
  }
  else {
    tilew = 4;
    tileh = 4;
  }

  k = 0;

  for (i = 0; i < h; i+=tileh) {

    for (j = 0; j < w; j+=tilew) {

      if (format == CARD_ICON_CI)
        GC_Card_TileC8(ptr, tlut, fb, x+(j/2), y+i, k++);
      else
        GC_Card_TileRGB(ptr, fb, x+(j/2), y+i, k++);
    }
  }

  return CARD_ERROR_NONE;
}
