// GCLIB (Free GameCube SDK Libraries)
//
// GC_EXI.c by Costis
//
// Version 1.0
//
// Revisions:
//
// V1.0 - Added all of the current functions, etc.
//		  Note: They are all synchronous and do not
//				use any interrupts.
//

#include <GC_EXI.h>

void GC_EXI_Init (void)
{
    // Wait until any previously active EXI operations have been completed on all of the EXI channels.
    while (EXI_DMACNT0 & EXI_CONTROL_ENABLE);
    while (EXI_DMACNT1 & EXI_CONTROL_ENABLE);
    while (EXI_DMACNT2 & EXI_CONTROL_ENABLE);

    // Clear the EXI Status Register for all of the EXI channels.
    EXI_STATUS0 = 0;
    EXI_STATUS1 = 0;
    EXI_STATUS2 = 0;
}

int GC_EXI_Probe (unsigned long channel)
{
	if (*(unsigned long*)(&EXI_STATUS0 + (channel * 5)) & 0x1000)
		return 1;
	else
		return 0;
}

// Channel must be from 0 to 2.
// Device must be from 0 to 2.
// Frequency must be from 0 to 5 (normally use 0).
void GC_EXI_Select (unsigned long channel, unsigned long device, unsigned long frequency)
{
    // Mask out the previously set device and frequency bits and set the new ones.
    *(unsigned long*)(&EXI_STATUS0 + (channel * 5)) = (*(unsigned long*)(&EXI_STATUS0 + channel * 5) & 0x405) |
                                                    (0x80 << device) | (frequency << 4) | 0x800;
}

// Channel must be from 0 to 2.
void GC_EXI_Deselect (unsigned long channel)
{
    // Mask out the previously set device and frequency bits.
    *(unsigned long*)(&EXI_STATUS0 + (channel * 5)) &= 0x405;
}

// Channel must be from 0 to 2.
// Size must be from 1 to 4.
// Type must be either EXI_CONTROL_TYPE_READ or EXI_CONTROL_TYPE_WRITE.
asm (".globl GC_EXI_Imm");
asm ("GC_EXI_Imm:");
asm ("mulli 3, 3, 20");
asm ("oris 3, 3, 0xCC00");
asm ("lwz 7, 0x6800(3)");
asm ("ori 7, 7, 0x8");
asm ("stw 7, 0x6800(3)");
asm ("subi 5, 5, 1");
asm ("slwi 7, 5, 4");
asm ("mtctr 6");
asm ("slwi 6, 6, 2");
asm ("or 6, 6, 7");
asm ("ori 6, 6, 1");
asm ("lwz 7, 0(4)");
asm ("stw 7, 0x6810(3)");
asm ("stw 6, 0x680C(3)");
asm ("GC_EXI_Imm_WaitDoneLoop:");
asm ("lwz 6, 0x680C(3)");
asm ("andi. 6, 6, 1");
asm ("bne GC_EXI_Imm_WaitDoneLoop");
asm ("bdz GC_EXI_Imm_SkipRead");
asm ("lwz 6, 0x6810(3)");
asm ("subfic 7, 5, 3");
asm ("slwi 7, 7, 3");
asm ("srw 6, 6, 7");
asm ("GC_EXI_Imm_ReadDataLoop:");
asm ("stbx 6, 4, 5");
asm ("srwi 6, 6, 8");
asm ("subic. 5, 5, 1");
asm ("bge GC_EXI_Imm_ReadDataLoop");
asm ("GC_EXI_Imm_SkipRead:");
asm ("blr");

void GC_EXI_Imma (unsigned long channel, unsigned char *abuffer, unsigned long size, unsigned long type)
{
    int i;
    unsigned long IMMdat;

	*(unsigned long*)(&EXI_STATUS0 + (channel * 5)) |= 0x8;

    if (type == EXI_CONTROL_TYPE_READ)
    {
		// EXI IMM Read Operation
		*(unsigned long*)(&EXI_DMACNT0 + (channel * 5)) = EXI_CONTROL_ENABLE | EXI_CONTROL_IMM |
																	  (type << 2) | ((size - 1) << 4);

		// Wait until the EXI IMM operation has been completed.
		while (*(volatile unsigned long*)(&EXI_DMACNT0 + (channel * 5)) & EXI_CONTROL_ENABLE);

		// Copy the EXI IMM data read to the buffer passed.
		IMMdat = *(unsigned long*)(&EXI_IMMDAT0 + (channel * 5)); 
		for (i = 0; i < size; i++)
			abuffer[i] = IMMdat >> ((3 - i) << 3);

		return;
    } else {
        // Copy the data from the buffer passed to the EXI IMM Data Register.
        IMMdat = 0;
        for (i = 0; i < size; i++)
            IMMdat |= abuffer[i] << ((3 - i) << 3);
        *(unsigned long*)(&EXI_IMMDAT0 + (channel * 5)) = IMMdat;

        // EXI IMM Write Operation
		*(unsigned long*)(&EXI_DMACNT0 + (channel * 5)) = EXI_CONTROL_ENABLE | EXI_CONTROL_IMM | 
                                                        (type << 2) | ((size - 1) << 4);
		// Wait until the EXI IMM operation has been completed.
		while (*(volatile unsigned long*)(&EXI_DMACNT0 + (channel * 5)) & EXI_CONTROL_ENABLE);

        return;
    }
}

void GC_EXI_ImmEx (unsigned long channel, unsigned char *abuffer, unsigned long size, unsigned long type)
{
	int i;

	if (size > 4)
	{
		for (i = 0; i < (size & ~3); i += 4)
			GC_EXI_Imm (channel, abuffer + i, 4, type);
		if (size & 3)
			GC_EXI_Imm (channel, abuffer + i, size & 3, type);
	} else
		GC_EXI_Imm (channel, abuffer, size, type);
}

// Channel must be from 0 to 2.
// Buffer must be aligned to a 32-bit offset.
// Size must be a multiple of 32 bytes.
// Type must be either EXI_CONTROL_TYPE_READ or EXI_CONTROL_TYPE_WRITE.
void GC_EXI_Dma (unsigned long channel, unsigned char *abuffer, unsigned long size, unsigned long type)
{
    // EXI DMA Operation
    *(unsigned long*)(&EXI_DMABUF0 + (channel * 5)) = (unsigned long)abuffer & 0x3FFFFE0;
    *(unsigned long*)(&EXI_DMALEN0 + (channel * 5)) = size & 0x3FFFFE0;
    *(unsigned long*)(&EXI_DMACNT0 + (channel * 5)) = EXI_CONTROL_ENABLE | EXI_CONTROL_DMA | (type << 2);

    // Wait until the EXI DMA operation has been completed.
    while (*(volatile unsigned long*)(&EXI_DMACNT0 + channel * 5) & EXI_CONTROL_ENABLE);

    return;
}

unsigned long GC_EXI_RetrieveID (unsigned long channel, unsigned long device)
{
	int i;
    unsigned long tID;

	if (GC_EXI_Probe (channel))
	{

		// Select the specified EXI channel and device.
		GC_EXI_Select (channel, device, 0);

		// Send the EXI ID command (0x0000)
		tID = 0;
		GC_EXI_Imm (channel, (unsigned char*)&tID, 2, EXI_CONTROL_TYPE_WRITE);

		// Read the actual ID data (4 bytes)
		GC_EXI_Imm (channel, (unsigned char*)&tID, 4, EXI_CONTROL_TYPE_READ);

		// Deselect the selected EXI device.
		GC_EXI_Deselect (channel);

		return tID;
	} else {
		return 0;
	}
}

char *GC_EXI_GetType (int type)
{
	if (type == EXI_MEMORY_CARD_59)
		return "Memory Card (59 blocks)";
	else if (type == EXI_MEMORY_CARD_123)
		return "Memory Card (123 blocks)";
	else if (type == EXI_MEMORY_CARD_251)
		return "Memory Card (251 blocks)";
	else if (type == EXI_MEMORY_CARD_507)
		return "Memory Card (507 blocks)";
	else if (type == EXI_MEMORY_CARD_1019)
		return "Memory Card (1019 blocks)";
	else if (type == EXI_MEMORY_CARD_2043)
		return "Memory Card (2043 blocks)";
	else if (type == EXI_USB_ADAPTER)
		return "USB Adapter";
	else if (type == EXI_IS_VIEWER)
		return "IS Viewer";
	else if (type == EXI_MODEM)
		return "Modem Adapter";
	else if (type == EXI_ETHER)
		return "BroadBand Adapter";
	else if (type == 0)
		return "Empty";
	else
		return "Unknown Device Type";
}

void GC_EXI_ReadROM (unsigned char *abuf, unsigned long offset, unsigned long size)
{
    unsigned long toffset;

    GC_EXI_Select (0, 1, 3);

    toffset = offset << 6;
    GC_EXI_Imm (0, (unsigned char*)&toffset, 4, 1);

    GC_EXI_Dma (0, abuf, size, 0);

	GC_EXI_Deselect (0);
}
