// GCLIB (Free GameCube SDK Libraries)
//
// GC_Interrupt.c by Costis
//
// Version 1.0
//
// Revisions:
//
// V1.0 - Added all of the current functions, etc.
//		  Some simple interrupt handling has been
//		  implemented.

#include <GC_Memory.h>
#include <GC_Context.h>
#include <GC_Exception.h>
#include <GC_AI.h>
#include <GC_EXI.h>
#include <GC_Interrupt.h>

typedef void *(*fp)(unsigned long, GC_Context*);

unsigned long *InterruptHandlerTable;

// This function is called in GC_System_Init.
// You should NOT have to call it directly.
void GC_Interrupts_Init (void)
{
	InterruptHandlerTable = (unsigned long *)0x80003040;

	// Clear the InterruptHandlerTable entries to 0.
	GC_Memory_memset ((char*)InterruptHandlerTable, 0, 128); // 32 interrupts

	INTERRUPT_CAUSE_CACHE = 0;	
	INTERRUPT_MASK = 0;

	*(unsigned long*)0xCC003004 = 0xF0; // PI STATUS

	// Initialize the interrupt mask. 
	//GC_Interrupts_MaskInternal (0xFFE0);
	
	//__GC_Exception_Handler_Set (4, (__GC_Exception_Handler)GC_Interrupts_ExternalInterruptHandler);

    //GC_Memory_memcpy((char*)0x80000500, (char*)GC_Interrupts_ExceptionVector, ((u32) GC_Interrupts_ExceptionVector_End - (u32) GC_Interrupts_ExceptionVector));
    //GC_Memory_DCFlushRangeNoSync((char*)0x80000500, GC_Interrupts_ExceptionVector_End - GC_Interrupts_ExceptionVector);
    //GC_Memory_ICInvalidateRange((char*)0x80000500, GC_Interrupts_ExceptionVector_End - GC_Interrupts_ExceptionVector);
}

// Some PowerPC assembly wrapper functions.
unsigned long cntlzw (unsigned long anum)
{
	asm ("cntlzw r3, r3");
	asm ("blr");
}

__inline void blrl (unsigned long address)
{
	asm ("blrl");
}

const unsigned long InterruptPriorityTable[ ] = {0x100, 0x40, 0xF8000000, 0x200,
											     0x80,  0x3000, 0x20, 0x3FF8C00,
											     0x4000000,  0x4000, 0xFFFFFFFF}; 


// GameCube reset button flag. This is set to 1 when the
// reset button has been pressed, if interrupts have been
// enabled. It is cleared to 0 by GC_System_Init.
unsigned long resetGC;

__GC_Exception_Handler GC_Interrupts_Dispatch (u32 exception, GC_Context *context)
{
	int i;
	u8 *inm, *outm;
	u32 interruptCause, interruptMask;
	u32 Icause;

	if (INTERRUPT_CAUSE & 0x00002)
	{
		resetGC = 1;
		INTERRUPT_MASK = 2; // Clear the interrupt mask!!!
	}

	if (INTERRUPT_CAUSE & 0x00000100)
		// Video Interface interrupt! Call the VI interrupt
		// handler, which includes a frame counter.
		GC_Video_InterruptHandler ();
	else if (INTERRUPT_CAUSE & 0x00000040) {
		// Audio Interface Interrupt. Useful for music players.
		// This code just clears the AI DMA interrupt flag.
		AI_DSP_STATUS = (AI_DSP_STATUS & 0xFFDF) | 0x8;
	}


	// Return from the interrupt handler.
	GC_Context_Load (context);


	// The code that follows is a generic interrupt handler,
	// that should support all interrupts rather than the
	// three handled by the specific one above. The generic
	// handler may not be completely ready for use, yet,
	// though, so replace the above code with the following
	// at your own risk!

	/*interruptCause = INTERRUPT_CAUSE & ~0x10000;

	interruptMask = *(unsigned long*)0xCC003004; // PI STATUS

	// Return from the function if there are no interrupts left to handle.
	//if ((!interruptCause) || (!(interruptCause & interruptMask)))
	//	GC_Context_Load (context);
	
	interruptMask = 0;
	
	if (interruptCause & 0x80)
	{
		// Memory Interface Interrupts
		Icause = *(u32*)(0xCC004030);
	
		if (Icause & 1)
			interruptMask |= 0x80000000;
		if (Icause & 2)
			interruptMask |= 0x40000000;
		if (Icause & 4)
			interruptMask |= 0x20000000;
		if (Icause & 8)
			interruptMask |= 0x10000000;
		if (Icause & 16)
			interruptMask |= 0x8000000;
	}

	if (interruptCause & 0x40)
	{
		// Audio Interface Interrupts
		Icause = *(u32*)(0xCC005010);

		if (Icause & 8)
			interruptMask |= 0x4000000;
		if (Icause & 32)
			interruptMask |= 0x2000000;
		if (Icause & 128)
			interruptMask |= 0x1000000;

		AI_DSP_STATUS = (AI_DSP_STATUS & 0xFFDF) | 0x8;
		GC_Context_Load (context);
	}

	if (interruptCause & 0x20)
	{
		// Streaming Interface Interrupts
		Icause = *(u32*)(0xCC006C00);

		if (Icause & 8)
			interruptMask |= 0x800000;
	}

	if (interruptCause & 0x10)
	{
		// Expansion Interface Interrupts

		// Channel 0
		Icause = *(u32*)(0xCC006800);

		if (Icause & 2)
			interruptMask |= 0x400000;
		if (Icause & 8)
			interruptMask |= 0x200000;
		if (Icause & 0x800)
			interruptMask |= 0x100000;

		// Channel 1
		Icause = *(u32*)(0xCC006814);

		if (Icause & 2)
			interruptMask |= 0x80000;
		if (Icause & 8)
			interruptMask |= 0x40000;
		if (Icause & 0x800)
			interruptMask |= 0x20000;

		// Channel 2
		Icause = *(u32*)(0xCC006828);

		if (Icause & 2)
			interruptMask |= 0x10000;
		if (Icause & 8)
			interruptMask |= 0x8000;
	}

	if (interruptCause & 0x2000)
		// HSP Interface Interrupts
		interruptMask |= 0x20;

	if (interruptCause & 0x1000)
		// Debug Interface Interrupts
		interruptMask |= 0x40;

	if (interruptCause & 0x400)
		// Error Interface Interrupts
		interruptMask |= 0x1000;

	if (interruptCause & 0x200)
		// HSP Interface Interrupts
		interruptMask |= 0x2000;

	if (interruptCause & 0x100)
	{
		// VI Interface Interrupts
		interruptMask |= 0x80;
		GC_Video_InterruptHandler ();
		GC_Context_Load (context);
	}

	if (interruptCause & 8)
		// HSP Interface Interrupts
		interruptMask |= 0x800;

	if (interruptCause & 4)
		// HSP Interface Interrupts
		interruptMask |= 0x400;

	if (interruptCause & 2)
	{
		// HSP Interface Interrupts
		interruptMask |= 0x200;
		resetGC = 1;
		GC_Context_Load ();
	}

	if (interruptCause & 0x800)
		// VI Interface Interrupts
		interruptMask |= 0x4000;

	if (interruptCause & 1)
		// VI Interface Interrupts
		interruptMask |= 0x100;

	if (!interruptMask)
		GC_Context_Load (context);

	i = 0;
	while (1)
	{
		if (interruptMask & InterruptPriorityTable[i])
		{
			interruptMask = cntlzw (interruptMask);
			break;
		} else i++;
	}

	__GC_Exception_Test (interruptMask, interruptCause, 0x12345678, 0x87654321);
	
	if (InterruptHandlerTable[interruptMask])
		((fp)(InterruptHandlerTable[interruptMask]))(interruptMask, context);

	GC_Context_Load (context);*/
}

	// PPC assembly code to handle the external interrupt exception!
	asm(".globl GC_Interrupts_ExternalInterruptHandler");
	asm("GC_Interrupts_ExternalInterruptHandler:");
	asm("stw      r0, 0(r4)");
	asm("stw      r1, 4(r4)");
	asm("stw      r2, 8(r4)");
	asm("stmw     r6, 24(r4)");
	asm("mfspr    r0, 913");
	asm("stw      r0, 424(r4)");
	asm("mfspr    r0, 914");
	asm("stw      r0, 428(r4)");
	asm("mfspr    r0, 915");
	asm("stw      r0, 432(r4)");
	asm("mfspr    r0, 916");
	asm("stw      r0, 436(r4)");
	asm("mfspr    r0, 917");
	asm("stw      r0, 440(r4)");
	asm("mfspr    r0, 918");
	asm("stw      r0, 444(r4)");
	asm("mfspr    r0, 919");
	asm("stw      r0, 448(r4)");
	asm("stwu     r1, -8(r1)");
	asm("mfmsr	  r3");
	asm("ori	  r3, r3, 0x2000");
	asm("mtmsr	  r3");
	asm("b        GC_Interrupts_Dispatch");

	// These functions must be written in PPC assembly
	// since they use the MSR register.
	asm (".globl GC_Interrupts_Disable");
	asm(".globl __GC_Context_InterruptsDisable_begin");
	asm(".globl __GC_Context_InterruptsDisable_end");
	asm ("GC_Interrupts_Disable:");
	asm ("__GC_Context_InterruptsDisable_begin:");
	asm ("mfmsr		r3");
	asm ("rlwinm	r4, r3, 0, 17, 15");
	asm ("mtmsr		r4");
	asm ("__GC_Context_InterruptsDisable_end:");
	asm ("extrwi	r3, r3, 1, 16");
	asm ("blr");

	asm (".globl GC_Interrupts_Enable");
	asm ("GC_Interrupts_Enable:");
	asm ("mfmsr		r3");
	asm ("ori		r4, r3, 0x8000");
	asm ("mtmsr		r4");
	asm ("extrwi	r3, r3, 1, 16");
	asm ("blr");

	asm (".globl GC_Interrupts_Restore");
	asm ("GC_Interrupts_Restore:");
	asm ("cmpwi    r3, 0");
	asm ("mfmsr    r4");
	asm ("beq      GC_Interrupts_Restore_None");
	asm ("ori      r5, r4, 0x8000");
	asm ("b        GC_Interrupts_Restore_NoneSkip");
	asm ("GC_Interrupts_Restore_None:");
	asm ("rlwinm   r5, r4, 0, 17, 15");
	asm ("GC_Interrupts_Restore_NoneSkip:");
	asm ("mtmsr    r5");
	asm ("extrwi   r3, r4, 1, 16");
	asm ("blr");
