// GCLIB (Free GameCube SDK Libraries)
//
// GC_Memory.c by Costis
//
// Version 1.0
//
// Revisions:
//
// V1.0 - Added all of the current functions, etc.
//		  Memory protection support has not been
//		  added yet, although that is not really
//		  a priority as of now. lint wrote the
//		  memory address conversion functions.

#include <GC_Memory.h>

void *GC_Memory_PhysicalToCached(u32 paddr)
{
#ifdef GC_DEBUG
	if(paddr >= 0x10000000)
	{
		printf("GC_Memory.c : GC_physical_to_cached : Illegal Address");
	}
#endif

	return (void*)(paddr + 0x80000000);
}

void *GC_Memory_PhysicalToUncached(u32 paddr)
{
#ifdef GC_DEBUG
	if(paddr >= 0x10000000)
	{
		printf("GC_Memory.c: GC_Memory_PhysicalToUncached: Illegal Address");
	}
#endif

	return (void*)(paddr + 0xC0000000);
}

u32 GC_Memory_CachedToPhysical(void *caddr)
{
#ifdef GC_DEBUG
	if(caddr < 0x80000000 || caddr >= 0x90000000)
	{
		printf("GC_Memory.c: GC_Memory_CachedToPhysical: Illegal Address");
		return;
	}
#endif

	return ((u32)caddr - 0x80000000);
}
	
u32 GC_Memory_UncachedToPhysical(void *ucaddr)
{
#ifdef GC_DEBUG
	if(ucaddr < 0xC0000000 || ucaddr >= 0xD0000000)
	{
		printf("GC_Memory.c: GC_Memory_UncachedToPhysical: Illegal Address"); 
		return;
	}
#endif

	return ((u32)ucaddr - 0xC0000000);
}

void *GC_Memory_CachedToUncached(void *caddr)
{
#ifdef GC_DEBUG	
	if(caddr < 0x80000000 || caddr >= 0x90000000)
	{
		printf("GC_Memory.c: GC_Memory_CachedToUncached: Illegal Address"); 
	}
#endif

	return (void*)((u32)caddr + 0x40000000);
}

void *GC_Memory_UncachedToCached(void *ucaddr)
{
#ifdef GC_DEBUG
	if(ucaddr < 0xC0000000 || ucaddr >= 0xD0000000)
	{
		printf("GC_Memory.c : GC_UncachedToCached: Illegal Address"); 
	}
#endif

	return (void*)((u32)ucaddr - 0x40000000);
}

extern void GC_Memory_ICFreeze (void);

	// Written in PPC assembly due to the SPR manipulation instructions.
	asm (".globl GC_Memory_ICFreeze");
	asm ("GC_Memory_ICFreeze:");
	asm ("isync");
	asm ("mfspr    r3, 1008");
	asm ("ori      r3, r3, 0x6000");
	asm ("mtspr    1008, r3");
	asm ("isync");

extern void GC_Memory_DCInvalidateRange (void* addr, u32 n);

	// Written in PPC assembly due to the cache invalidation instructions.
	// r3 = address (should be aligned to 32 bits)
	// r4 = size (should be a multiple of 32)

	asm (".globl GC_Memory_DCInvalidateRange");
	asm ("GC_Memory_DCInvalidateRange:");
	asm ("cmplwi   r4,0x0000");
	asm ("blelr");
	asm ("clrlwi   r5,r3,27");
	asm ("add      r4,r4,r5");
	asm ("addi     r4,r4,31");
	asm ("srwi     r4,r4,5");
	asm ("mtctr    r4");
	asm ("GC_Memory_DCInvalidateRange_Loop:");
	asm ("dcbi     r0,r3");
	asm ("addi     r3,r3,32");
	asm ("bdnz     GC_Memory_DCInvalidateRange_Loop");
	asm ("blr");

extern void GC_Memory_DCZeroRange (void* addr, u32 n);

	asm(".globl GC_Memory_DCZeroRange");

	// Written in PPC assembly due to the cache zeroing instructions.
	// r3 = address (should be aligned to 32 bits)
	// r4 = size (should be a multiple of 32)

	asm("GC_Memory_DCZeroRange:");
	asm ("cmplwi   r4, 0");
	asm ("blelr");
	asm ("clrlwi   r5, r3, 27");
	asm ("add      r4, r4, r5");
	asm ("addi     r4, r4, 31");
	asm ("srwi     r4, r4, 5");
	asm ("mtctr    r4");
	asm ("GC_Memory_DCZeroRange_Loop:");
	asm ("dcbz     r0, r3");
	asm ("addi     r3, r3, 32");
	asm ("bdnz     GC_Memory_DCZeroRange_Loop");
	asm ("blr");

extern void GC_Memory_DCFlushRange (void* addr, u32 n);

	asm(".globl GC_Memory_DCFlushRange");

	// Written in PPC assembly due to the cache flushing instructions.
	// r3 = address (should be aligned to 32 bits)
	// r4 = size (should be a multiple of 32)

	asm("GC_Memory_DCFlushRange:");
	asm ("cmplwi   r4, 0");
	asm ("blelr");
	asm ("clrlwi   r5, r3, 27");
	asm ("add      r4, r4, r5");
	asm ("addi     r4, r4, 31");
	asm ("srwi     r4, r4, 5");
	asm ("mtctr    r4");
	asm ("GC_Memory_DCFlushRange_Loop:");
	asm ("dcbf     r0, r3");
	asm ("addi     r3, r3, 32");
	asm ("bdnz     GC_Memory_DCFlushRange_Loop");
	asm ("sc");
	//asm ("mfspr    r3, 1008");
	//asm ("ori      r4, r3,0x0008");
	//asm ("mtspr    1008, r4");
	//asm ("isync");
	//asm ("sync");
	//asm ("mtspr    1008, r3");
	asm ("blr");

extern void GC_Memory_DCFlushRangeNoSync (void* addr, u32 n);

	asm(".globl GC_Memory_DCFlushRangeNoSync");

	// Written in PPC assembly due to the cache flushing instructions.
	// r3 = address (should be aligned to 32 bits)
	// r4 = size (should be a multiple of 32)

	asm("GC_Memory_DCFlushRangeNoSync:");
	asm("cmplwi   r4, 0"); // Check if the size is 0.
	asm("blelr");          // If so, exit the function.
	asm("clrlwi.  r5,r3,27"); // Check if the address is aligned to 32 bytes.
	asm("beq      GC_Memory_DCFlushRangeNoSync_AddrAligned");
	asm("addi     r4,r4,32");
	asm("GC_Memory_DCFlushRangeNoSync_AddrAligned:");
	asm("addi     r4,r4,31");
	asm("srwi     r4,r4,5");
	asm("mtctr    r4"); // Set the size to the counter register.
	asm("GC_Memory_DCFlushRangeNoSync_FlushLoop:");
	asm("dcbf     r0,r3"); // Data Cache flushing loop.
	asm("addi     r3,r3,32"); // Increase the address by 32 bytes.
	asm("bdnz     GC_Memory_DCFlushRangeNoSync_FlushLoop");
	asm("blr");

extern void GC_Memory_ICInvalidateRange (void* addr, u32 n);

	asm(".globl GC_Memory_ICInvalidateRange");
	
	// Written in PPC assembly due to the cache invalidation instructions.
	// r3 = address (should be aligned to 32 bits)
	// r4 = size (should be a multiple of 32)
	asm("GC_Memory_ICInvalidateRange:");
    asm("cmplwi   r4,0x0000"); // Check if the size is 0.
    asm("blelr"); // If so, then exit the function.
    asm("clrlwi.  r5,r3,27"); // Check if the address is aligned to 32 bytes.
    asm("beq      GC_Memory_ICInvalidateRange_AddrAligned");
    asm("addi     r4,r4,32");
	asm("GC_Memory_ICInvalidateRange_AddrAligned:");
    asm("addi     r4,r4,31");
    asm("srwi     r4,r4,5");
    asm("mtctr    r4"); // Set the size to the counter register.
    asm("icbi     r0,r3"); // Instruction Cache invalidation loop.
    asm("addi     r3,r3,32"); // Increase the address by 32 bytes.
    asm("bdnz     $-8");
    asm("sync");
    asm("isync");
    asm("blr");

void GC_Memory_memcpy (char *dest, char *src, u32 size)
{
	int i;

	// Could be optimized later.
	for (i = 0; i < size; i++)
		*dest++ = *src++;
}

void GC_Memory_memset (char *dest, int anum, u32 size)
{
	int i;

	// Could be optimized later.
	for (i = 0; i < size; i++)
		*dest++ = anum;
}

/*
  Just taken from libc source.
*/

int GC_Memory_memcmp(const void *av, const void *bv, int len) {

  const unsigned char *a = av;
  const unsigned char *b = bv;
  int i;

  for (i=0; i<len; i++) {
    if (a[i] != b[i]) {
      return (int)(a[i] - b[i]);
    }
  }

  return 0;
}